package org.tsutils;



/**
 * To generate header: javah -classpath target/classes -jni -d src/main/c org.tsutils.RdtscTimestamper 
 * 
 * To run: java -cp target/classes -Djava.library.path=src/main/resources/linux org.tsutils.RdtscTimestamper
 * 
 * Based on Peter Lawrey code: https://github.com/peter-lawrey/Java-Thread-Affinity/blob/master/src/main/java/vanilla/java/clock/impl/JNIClock.java
 * 
 * @author Sergio Oliveira Jr.
 */
public class RdtscTimestamper extends AbstractTimestamper {
	
    private static final int FACTOR_BITS = 17;

    private final long start;
    private long factorValue;
    
    public RdtscTimestamper() {
    	super();
    	estimateFrequency(50);
    	estimateFrequency(200);
    	this.start = rdtsc();
    }
	
    private final long tscToNano(final long tsc) {
        return (tsc * factorValue) >> FACTOR_BITS;
    }
    
	private final native long rdtsc();
	
	@Override
    public final long nanoTime() {
        return tscToNano(rdtsc() - start);
    }
	
	static {
		System.loadLibrary(RdtscTimestamper.class.getSimpleName());
	}
	
    private void estimateFrequency(int factor) {
    	final long start = System.nanoTime();
        long now;
        while ((now = System.nanoTime()) == start);
        long end = start + factor * 1000000L;
        final long start0 = rdtsc();
        while ((now = System.nanoTime()) < end);
        long end0 = rdtsc();
        end = now;

        factorValue = ((end - start) << FACTOR_BITS) / (end0 - start0) - 1;
    }
	
	public static void main(String[] args) {
		Timestamper ts = new RdtscTimestamper();
		System.out.println(ts.nanoTime());
		System.out.println(ts.now());
	}
}